import asyncdispatch, bcrypt, jester, json, jwt, os, osproc, sequtils, tempfile,
  times, unittest

import ./configuration, ./core, private/util

type Worker = object
  process*: Process
  workingDir*: string

type
  Session = object
    user*: UserRef
    issuedAt*, expires*: Time

const ISO_TIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss"

proc makeJsonResp(status: HttpCode, details: string = ""): string =
  result = $(%* {
    "statusCode": status.int,
    "status": $status,
    "details": details
  })

proc newSession*(user: UserRef): Session =
  result = Session(
    user: user,
    issuedAt: getTime(),
    expires: daysForward(7).toTime())

proc toJWT*(cfg: StrawBossConfig, session: Session): string =
#  result = toJWT(%* {
#    "header": {
#      "alg": "HS256",
#      "typ": "JWT" },
#    "claims": {
#      "sub": session.user.name,
#      "iat": session.issuedAt.toSeconds().int,
#      "exp": session.expires.toSeconds().int } })

  var jwt = JWT(
    header: JOSEHeader(alg: HS256, typ: "jwt"),
    claims: toClaims(%*{
      "sub": session.user.name,
      "iat": session.issuedAt.toSeconds().int,
      "exp": session.expires.toSeconds().int }))

  jwt.sign(cfg.authSecret)
  result = $jwt

proc fromJWT*(cfg: StrawBossConfig, strTok: string): Session = 
  let jwt = toJWT(strTok)
  var secret = cfg.authSecret
  if not jwt.verify(secret): raiseEx "Unable to verify auth token."
  jwt.verifyTimeClaims()

  # Find the user record (if authenticated)
  let username = jwt.claims["sub"].node.str
  let users = cfg.users.filterIt(it.name == username)
  if users.len != 1: raiseEx "Could not find session user."

  result = Session(
    user: users[0],
    issuedAt: fromSeconds(jwt.claims["iat"].node.num),
    expires: fromSeconds(jwt.claims["exp"].node.num))

proc extractSession(cfg: StrawBossConfig, request: Request): Session =

  # Find the auth header
  if not request.headers.hasKey("Authentication"):
    raiseEx "No auth token."

  # Read and verify the JWT token
  result = fromJWT(cfg, request.headers["Authentication"])

proc spawnWorker(req: RunRequest): Worker =
  let dir = mkdtemp()
  var args = @["run", req.projectName, req.stepName, "-r", req.buildRef, "-w", dir]
  if req.forceRebuild: args.add("-f")
  result = Worker(
    process: startProcess("strawboss", ".", args, loadEnv(), {poUsePath}),
    workingDir: dir)

proc hashPwd*(pwd: string, cost: int8): string =
  let salt = genSalt(cost)
  result = hash(pwd, salt)

proc validatePwd*(u: UserRef, givenPwd: string): bool =
  let salt = u.hashedPwd[0..28] # TODO: magic numbers
  result = compare(u.hashedPwd, hash(givenPwd, salt))

proc makeAuthToken*(cfg: StrawBossConfig, uname, pwd: string): string =
  if uname == nil or pwd == nil:
    raiseEx "fields 'username' and 'password' required"

  # find the user record
  let users = cfg.users.filterIt(it.name == uname)
  if users.len != 1: raiseEx "invalid username or password"

  let user = users[0]

  if not validatePwd(user, pwd): raiseEx "invalid username or password"
  result = toJWT(cfg, newSession(user))

template requireAuth() =
  var session {.inject.}: Session
  try: session = extractSession(givenCfg, request)
  except: resp(Http401, makeJsonResp(Http401), "application/json")

proc start*(givenCfg: StrawBossConfig): void =

  let stopFuture = newFuture[void]()
  var workers: seq[Worker] = @[]

  settings:
    port = Port(8180)
    appName = "/api"

  routes:
    get "/ping":
      resp($(%*"pong"), "application/json")

    get "/auth-token":
      echo $request.params
      try:
        let authToken = makeAuthToken(givenCfg, @"username", @"password")
        resp("\"" & $authToken & "\"", "application/json")
      except: resp(Http401, makeJsonResp(Http401, getCurrentExceptionMsg()))

    get "/projects":
      requireAuth()
      resp($(%(givenCfg.projects)), "application/json")

    post "/project/@projectName/@stepName/run/@buildRef?":
      workers.add(spawnWorker(RunRequest(
        projectName: @"projectName",
        stepName: @"stepName",
        buildRef: if @"buildRef" != "": @"buildRef" else: nil,
        forceRebuild: false)))  # TODO support this with optional query params

    post "/service/debug/stop":
      if not givenCfg.debug: resp(Http404, makeJsonResp(Http404), "application/json")
      else:
        callSoon(proc(): void = complete(stopFuture))
        resp($(%*"shutting down"), "application/json")

  waitFor(stopFuture)
